######################################################################
## The functions are arranged into several groups:

## MAJOR functions that drive most of the work.
##   implementModel
##   getPredictionsOnly
##   prepare
##   wrapper
##   predictUnanimous
##   predictMixed
##   bestNonBuggyTree

## VARIABLE NAME functions that return lists of variable names for various purposes
##   getJDirVarNames
##   getJPredVarNames
##   getPredictors
##   getJusticeNames
##   getIDVarNames
##   getFileVarNames

## PREP AUX functions that help with preparing the data files for use
##   combinePetitResp
##   convertCircuit
##   convertEntity
##   spaethValue

## COUNTING functions that summarize information created by some of the other functions
##   sumVotes
##   justiceMisclassCount
##   justiceMisclassCountDF
##   getCourtdir
##   misclass.rate
##   newjMat

## FORMULA SPECIFIC functions that create and manipulate formulas passed to trees
##   baseform
##   justform
##   add.term

## AUXILIARY functions that do other small tasks
##   VoteCode
##   checkpoint
##   flush


######################################################################
##
## MAJOR functions start here
##
implementModel <- function (train, test, predsonly=TRUE) {
  if(missing(train)) train <- "fcast0930.csv"
  if(missing(test)) test <- "newtestdata.csv"
  predvars <- getJPredVarNames()
  if(any(predvars %in% ls(envir=parent.frame()))) stop ("Remove justice pred vars from toplevel environment before running.")
  input <- read.csv(train)
  names(input) <- tolower(names(input))
  newcsv <- read.csv(test)
  names(newcsv) <- tolower(names(newcsv))
  traincases <- prepare(input, condenseEntities = 1, combineEntities = 0)
  newcases <- prepare(newcsv, condenseEntities=1, combineEntities = 0, datatype="newdata")
  
  wdata <- wrapper(traincases, newcases, condenseEntities=1, combineEntities=0, criterion="minRate")
  
  if(predsonly) wdata <- getPredictionsOnly(wdata)
  
  return(wdata)
}

getPredictionsOnly <- function(idata) {
  # idata is expected to be an object of the type returned by wrapper()
  cases <- idata$wholetesting
  summs <- idata$wholetestout
  allofem <- merge(cases[order(cases$docket),], summs[order(summs$docket),], by="docket")
  predvars <- getJPredVarNames()
  keepnames <- c("docket", predvars, "uConPred", "uLibPred", "jConPred", "jLibPred", "cDirPred", "lctdir")
  extract <- allofem[,which(names(allofem) %in% keepnames)]
  return(extract)
}

prepare <- function(df, combineEntities=c(0,1), condenseEntities=c(0,1), datatype=c("hasdir", "newdata"), ...) {
  
  datatype <- "hasdir"
  
  cat("prepare(", deparse(substitute(df)), ")...\n", sep="")
  justnames <- getJusticeNames(...)
  idvars <- getIDVarNames(...)
  dirvars <- getJDirVarNames(...)
  reqdvars <- getFileVarNames(combineEntities, ...)

  ## Is everything we need here?  Stop if not.
  if(datatype == "hasdir") {
    getcols <- c(idvars, dirvars, reqdvars)
  } else {
    getcols <- c(idvars, reqdvars)
  }
  keepcols <- which(names(df) %in% getcols)  
  if(length(keepcols) != length(getcols)) {
    stop(paste("At least one of the following variables is missing from the input data frame:\n",
      setdiff(getcols, names(df))))
  }
  
  ## Establish return dataframe...
  keepcols <- which(names(df) %in% c(idvars, reqdvars))
  retdf <- df[,keepcols]
  
  ## Convert variables to factors, assigning numbers...
  retdf$lctdir <- factor(df$lctdir, levels=c("0", "1"))
  retdf$circnum <- convertCircuit(df$circuit, ...)
  retdf$petitnum <- convertEntity(df$petit, condenseEntities, ...)
  retdf$respnum <- convertEntity(df$respon, condenseEntities, ...)
  retdf$prcode <- as.factor(combinePetitResp(retdf$petitnum, retdf$respnum, ...))
  
  ## Justice direction variables: convert to factors...
  attach(df)
  if(datatype == "hasdir") {
    ## There's gotta be a better way to do this... but it works...
    retdf$rehndir <- factor(rehndir, levels=c("0", "1"))
    retdf$stevdir <- factor(stevdir, levels=c("0", "1"))
    retdf$ocondir <- factor(ocondir, levels=c("0", "1"))
    retdf$scaldir <- factor(scaldir, levels=c("0", "1"))
    retdf$ kendir <- factor( kendir, levels=c("0", "1"))
    retdf$soutdir <- factor(soutdir, levels=c("0", "1"))
    retdf$thomdir <- factor(thomdir, levels=c("0", "1"))
    retdf$ gindir <- factor( gindir, levels=c("0", "1"))
    retdf$ brydir <- factor( brydir, levels=c("0", "1"))
  } else {
    retdf$rehndir <- rep(-1, nrow(df))
    retdf$stevdir <- rep(-1, nrow(df))
    retdf$ocondir <- rep(-1, nrow(df))
    retdf$scaldir <- rep(-1, nrow(df))
    retdf$ kendir <- rep(-1, nrow(df))
    retdf$soutdir <- rep(-1, nrow(df))
    retdf$thomdir <- rep(-1, nrow(df))
    retdf$ gindir <- rep(-1, nrow(df))
    retdf$ brydir <- rep(-1, nrow(df))
  }
  
  if(is.factor(issue <- df$issue)) {
    checkpoint("Variable 'issue' is a factor")
    if(!(0 %in% levels(issue))) levels(issue) <- c(levels(issue), 0)
    issue[issue=="."] <- 0
    retdf$value <- spaethValue(as.integer(as.character(issue)), ...)
  } else {
    issue[is.na(issue)] <- 0
    retdf$value <- spaethValue(issue, ...)
  }

  retdf$jur <- factor(jur, levels=c(as.character(1:3)))
  retdf$lcdisag <- factor(lcdisag, levels=c('0','1'))
  retdf$statesc <- factor(statesc, levels=c('0','1'))
  retdf$unconst <- factor(unconst, levels=c('0','1'))
  retdf$otprec <- factor(otprec, levels=c('0','1'))

  rsv <- sumVotes(retdf, dirvars, ...)
  tossrows <- which(is.na(rsv$Ones) & is.na(rsv$Zeros)) 

  if(datatype == "hasdir") {
 
    if(length(tossrows) > 0) {  
      cat(paste("Removing",length(tossrows),"cases with all justice direction variables == NA:\n"))
      cat(tossrows, "\n")
      retdf <- retdf[-tossrows, ]
      rsv <- rsv[-tossrows, ]
    }
  }
  ## 1 = Conservative, 0 = Liberal
  retdf$unanCon <- as.integer(rsv$Zeros==0 & rsv$Ones > 0)
  retdf$unanLib <- as.integer(rsv$Zeros > 0 & rsv$Ones==0)
  attach(retdf)
  retdf$unanimous <- as.integer(unanCon | unanLib)
  retdf$voteType <- as.factor(ifelse(unanCon, 2, ifelse(unanLib, 0, 1)))

  return(retdf)
}

wrapper <- function(df.train, df.test, ...) {
  ## Takes df.train, builds trees and applies them to df.test
  cat("\nTraining data has ", nrow(df.train), " rows and testing data has ", nrow(df.test))
  
  treeDepVars <- c("unanCon", "unanLib", getJDirVarNames(...))
  testFrameMissing <- setdiff(treeDepVars, names(df.test))
  if(length(testFrameMissing) > 0) {
    for(varstring in testFrameMissing) {
      oldnames <- names(df.test)
      newvar <- rep(-1, nrow(df.test))
      df.test <- cbind(df.test, newvar)
      names(df.test) <- c(oldnames, varstring)
    }
  }

  udata <- predictUnanimous(df.train, df.test, ...)
  utesting <- udata$testing
  utrees <- udata$training.trees
  keeprows <- which(as.integer(udata$testout$uConPred) == as.integer(udata$testout$uLibPred))
  ##cat("\nKeeping rows ", keeprows, " from training data to send to predictMixed")
  mtesting <- udata$testing[keeprows, -which(names(udata$testing) %in% getJPredVarNames())]
  mdata <- predictMixed(df.train, mtesting, ...)
  mtesting <- mdata$testing
  mtrees <- mdata$training.trees
  
  ## checkpoint("\nAssigning wholetesting...")
  wholetesting <- rbind(utesting[-keeprows, ], mtesting)
  ## checkpoint("\nAssigning wholetestout...")
  wholetestout <- rbind(udata$testout[-keeprows, ], mdata$testout)

  ## checkpoint("\nAssigning retObject in wrapper()...")
  retObject <- list(wholetesting, wholetestout, utrees, mtrees)
  ## checkpoint("\nAssigning names(retObject) in wrapper()...")
  names(retObject) <- c("wholetesting", "wholetestout", "utrees", "mtrees")
  return(retObject)
}


predictUnanimous <- function(df.train, temp.test, percentage=90, predict=c("newdata", "testdata"), ...) {
  require(tree)
  # Takes df (a dataframe) and returns an object with predicted unanimous values
  # for each of the cases in df.train and df.test.  If df.test is named, all of df.train
  # is used to build the trees.  Otherwise, a random amount of df.train is selected,
  # based on percentage.

  predict <- match.arg(predict)
  
  df.test <- NA
  if(missing(temp.test) & predict == "newdata") {
    stop("Data frame must be passed to predictUnanimous() for predict='newdata'")
  }
  if(missing(temp.test)) {
    ## cat("\n\nTaking new random sample of ", percentage, "% of data\n"); flush()
    keeprows <- sample(1:nrow(df.train), nrow(df.train)*percentage/100)
    df.test <- df.train[-keeprows, ]
    df.train <- df.train[keeprows, ]
  } else {
    df.test <- temp.test
  }

  predictors <- getPredictors(...)
  unandirs <- c("unanCon", "unanLib")
  jPredVars <- getJPredVarNames(...)
  
  cat("\nChecking that the right variables exist...");flush()
  if(length(which(predictors %in% names(df.train))) < length(predictors)) {
    stop(paste("\n\npredictUnanimous: Training data set is missing\n", setdiff(predictors, names(df.train))))
  }
  if(length(which(predictors %in% names(df.test)))  < length(predictors)) {
    stop(paste("\n\npredictUnanimous: Testing data set is missing\n", setdiff(predictors, names(df.test))))
  }
  
  if(length(which(unandirs %in% names(df.train)))  < length(unandirs)) {
    stop(paste("\n\npredictUnanimous: Training data set is missing\n", setdiff(unandirs, names(df.train))))
  }
  if(predict=="testdata") {
    if(length(which(unandirs %in% names(df.test)))  < length(unandirs)) {
      stop(paste("\n\npredictUnanimous: Testing data set is missing\n", setdiff(unandirs, names(df.test))))
    }
  }
  
##  cat("\nCreating formulas for trees...");flush()
  attach(df.train)
  startform <- baseform(...)
  formCon <- justform(startform, "as.factor(unanCon)")
  formLib <- justform(startform, "as.factor(unanLib)")
  formulae <- list(formCon, formLib)

##  cat("\nFinding trees from training sample...");flush()
  treeCon <- bestNonBuggyTree(formCon)
  treeLib <- bestNonBuggyTree(formLib)
  training.trees <- list(treeCon, treeLib)
  
  predCon <- treeCon$frame$yval[treeCon$where]
  predLib <- treeLib$frame$yval[treeLib$where]
  
##  cat("\nAdding variables for justice predicted votes to training dataframe...");flush()
  votes <- VoteCode()
  unanPred <- rep(0.5, nrow(df.train))
  unanPred[predCon == 1 & predLib == 0] <- votes$Conservative
  unanPred[predCon == 0 & predLib == 1] <- votes$Liberal
  oldnames <- names(df.train)
  df.train$unanPred <- unanPred
  names(df.train) <- c(oldnames, "unanPred")
  for(pred in jPredVars) {
    oldnames <- names(df.train)
    tempmat <- cbind(as.matrix(df.train), ifelse((unanPred==0 | unanPred==1),unanPred,NA))
    df.train <- data.frame(tempmat)
    names(df.train) <- c(oldnames, pred)
  }

## checkpoint("Before attaching df.test")  
  ## cat("\nUsing trees built on training data to predict responses in testing data...");flush()

  attach(df.test)

## checkpoint(paste("df.test has", nrow(df.test), "rows"))
  treeConPred <- predict.tree(treeCon, df.test, type="tree")
    treeConPred$weights <- rep(1, nrow(df.test))  ## KLUDGE
  treeConPred$y <- rep(NA, nrow(df.test))
  if("unanCon" %in% names(df.test)) treeConPred$y <- df.test[, which(names(df.test) == "unanCon")]
  
## checkpoint(paste("df.test has", nrow(df.test), "rows"))
  treeLibPred <- predict.tree(treeLib, df.test, type="tree")
    treeLibPred$weights <- rep(1, nrow(df.test))  ## KLUDGE
  treeLibPred$y <- rep(NA, nrow(df.test))
  if("unanLib" %in% names(df.test)) treeLibPred$y <- df.test[, which(names(df.test) == "unanLib")]
  testing.trees <- list(treeConPred, treeLibPred)
  
  predCon <- treeConPred$frame$yval[treeConPred$where]
  predLib <- treeLibPred$frame$yval[treeLibPred$where]
  
  ## cat("\nBuilding variables for justice predicted votes in testing data...");flush()
  ## checkpoint(paste("df.test has", nrow(df.test), "rows"))
  votes <- VoteCode()
  uPred <- rep(0.5, nrow(df.test))
  uPred[predCon == 1 & predLib == 0] <- votes$Conservative
  uPred[predCon == 0 & predLib == 1] <- votes$Liberal
  oldnames <- names(df.test)
  df.test$unanPred <- uPred
  colnames(df.test) <- c(oldnames, "unanPred")
  for(pred in jPredVars) {
    oldnames <- colnames(df.test)
    tempmat <- cbind(as.matrix(df.test), ifelse((uPred==0 | uPred==1),uPred,NA))
    df.test <- data.frame(tempmat)
    colnames(df.test) <- c(oldnames, pred)
  }
  
## checkpoint("Before creating misclassrates dataframe")
  ## cat("\nCalculating misclassification rates...");flush()
  misclassrates <- data.frame(matrix(0, nr=2, nc=2))
  if(predict=="testdata") {
    colnames(misclassrates) <- c("Train", "Test")
    row.names(misclassrates) <- c("Con", "Lib")
    rCon <- which(row.names(misclassrates) == "Con")
    rLib <- which(row.names(misclassrates)== "Lib")
    cTrain <- which(colnames(misclassrates) == "Train")
    cTest <- which(colnames(misclassrates) == "Test")
    misclassrates[rCon, cTrain] <- misclass.rate(treeCon)
    misclassrates[rLib, cTrain] <- misclass.rate(treeLib)
    misclassrates[rCon, cTest] <- misclass.rate(treeConPred)
    misclassrates[rLib, cTest] <- misclass.rate(treeLibPred)
  }
 
## checkpoint("Before assigning new columns to df.test")
  ## cat("\nCreating testout object...");flush()
  df.test$unanLibPred <- as.integer(as.character(treeLibPred$frame$yval[treeLibPred$where]))
  df.test$unanConPred <- as.integer(as.character(treeConPred$frame$yval[treeConPred$where]))

  attach(df.test)
  justdirs <- getJDirVarNames()
  justpreds <- getJPredVarNames()

  docket <- as.character(df.test$docket)
  uLibPred <- df.test$unanLibPred
  uConPred <- df.test$unanConPred

  if(predict == "testdata") {
    sVotes <- sumVotes(df.test, justdirs)
    jLibAct <- sVotes$Zeros
    jConAct <- sVotes$Ones
    cDirAct <- getCourtdir(df.test, method="dir")
  } else {
    jLibAct <- rep(NA, nrow(df.test))
    jConAct <- rep(NA, nrow(df.test))
    cDirAct <- rep(NA, nrow(df.test))
  }

  sVotes <- sumVotes(df.test, justpreds)
  jLibPred <- sVotes$Zeros
  jConPred <- sVotes$Ones
  cDirPred <- getCourtdir(df.test, method="pred")
  
  if(predict == "testdata") {
    jMiscount <- justiceMisclassCountDF(df.test)
    cMisclass <- as.integer(cDirPred != cDirAct)
  } else {
    jMiscount <- rep(NA, nrow(df.test))
    cMisclass <- rep(NA, nrow(df.test))
  }
  
  testout <- as.data.frame(cbind(docket, uConPred, uLibPred, jConAct, jConPred, jLibAct, jLibPred, cDirAct, cDirPred, jMiscount, cMisclass))

  ## cat("\nCreating object returned by predictedUnanimous()");flush()
  retObject <- list(justdirs, df.train, df.test, misclassrates, formulae, 
    training.trees, testing.trees, testout)
  names(retObject) <- c("justnames", "training", "testing", "misclassrates", 
    "formulae", "training.trees", "testing.trees", "testout")
  
  cat("\n")
  return(retObject)
  
}

predictMixed <- function(df.train, temp.test=NULL, percentage=90, quiet=TRUE, criterion=c("PRE", "minRate"),
  method=c("predicted", "actual"), random=c("new", "static"), predict=c("newdata", "testdata"), ...) {
  require(tree)
  ## We check for the existence of all necessary variables

  predict <- match.arg(predict)
  if(predict == "newdata" & missing(temp.test)) {
    stop("Data frame must be supplied to predictMixed() for option predict='newdata'")
  }

  df.test <- NA
  if(missing(temp.test)) {
##    cat("\n\nTaking new random sample of ", percentage, "% of data\n"); flush()
    keeprows <- sample(1:nrow(df.train), nrow(df.train)*percentage/100)
    df.test <- df.train[-keeprows, ]
    df.train <- df.train[keeprows, ]
  } else {
    df.test <- temp.test
  }
  
  predictors <- getPredictors(...)
  justnames <- getJusticeNames(...)
  jDirVars <- getJDirVarNames(...)
  jPredVars <- getJPredVarNames(...)

  criterion <- match.arg(criterion)
  method <- match.arg(method)  
  random <- match.arg(random)
  
  ## percentage limited to range from 10 to 100
  percentage <- max(min(100, percentage), 10)
    
  if(criterion == "PRE") {
    marginalPct <- newjMat(df.train)$jBenchmark
  } else {
    marginalPct <- rep(0.5, length(justnames))
  }
  names(marginalPct) <- paste(justnames, "marg", sep="")
  
## cat("\nVerifying variables existence...")
  if(length(which(jDirVars %in% names(df.train))) != length(jDirVars)) {
    stop(paste("\n\npredictMixed(): dataframe doesn't contain ", setdiff(jDirVars, names(df.train)), "\n",sep=""))
  }
  if(!quiet) cat("\n\npredictMixed(): Justice direction variables all present and accounted for.")

  jDirMat <- df.train[,which(names(df.train) %in% jDirVars)]
  if(!quiet) { cat("\nnames(jDirMat): "); cat(colnames(jDirMat)); flush() }

  if(length(which(predictors %in% names(df.train))) != length(predictors)) {
    stop(paste("\n\npredictMixed(): dataframe doesn't contain ", setdiff(predictors, names(df.train)), "\n", sep=""))
  }

  if(!missing(df.test)) {
    if(length(which(predictors %in% names(df.test))) != length(predictors)) {
      stop(paste("\n\npredictMixed(): df.test doesn't contain ", setdiff(predictors, names(df.test)), "\n", sep=""))
    }
  }
  
  if(!quiet) cat("\n\npredictMixed(): Predictor variables all present and accounted for.")
  predcols <- df.train[, which(names(df.train) %in% predictors)]
  if(!quiet) { cat("\nnames(predcols): "); cat(colnames(predcols)); flush() }
  
  if(!("docket" %in% names(df.train))) { docket <- 1:nrow(jDirMat)
  } else { docket <- as.character(df.train$docket) }
  alldata <- data.frame(jDirMat, predcols)
  row.names(alldata) <- docket

  if(percentage == 100 | !missing(df.test)) {
    ## ...that is, we use all the rows of df to build the models if:
    ## 1. We're told to use all the rows, OR
    ## 2. There is a test data set identified.
    keeprows <- 1:nrow(df.train)
    if(!missing(df.test)) {
      ## cat("\nUsing all rows of training data to predict responses for new data")
    }
  } else {
    if(random == "new" | !("randunif" %in% names(df.train))) 
    { ## cat("\n\nTaking new random sample of ", percentage, "% of data"); flush()
      keeprows <- sample(1:nrow(df.train), nrow(df.train)*percentage/100)
    } else {
      randunif <- df$randunif
      if(any(is.na(randunif))) {
        ## cat("\n\nAt least some randunif values from dataframe are missing; generating new ones"); flush()
        randunif[is.na(randunif)] <- runif(length(randunif[is.na(randunif)]))
      }
      keeprows <- row(as.matrix(randunif[randunif <= (percentage/100)]))
    }
  }

  training <- alldata[keeprows, ]
  if(missing(df.test)) {
    if(percentage < 100) {
      testing <- alldata[-keeprows, ]
    } else {
      testing <- training
    }
  } else {
    testing <- df.test
  }
  
  ## cat("\n"); flush()

  baseJust <- c(1:length(jDirVars))
  basePreds <- c(length(jDirVars)+1:length(jDirVars)+length(predictors))
    
  attach(training)
  startform <- baseform(...)
  forms <- justform(startform, jDirVars[1])
  for(i in 2:length(jDirVars)) {
    forms <- c(forms, justform(startform, jDirVars[i]))
  }
  
  swap <- rep(NA, ncol(training))
  misclassrates <- matrix(NA, nr=length(baseJust), nc=length(baseJust))
  
  offset <- 0
  if(method=="predicted") {
    offset <- ncol(training)		## Used later
  }
  predvars <- as.data.frame(matrix(NA, nr=nrow(training), nc=9))
  names(predvars) <- paste(justnames, "pred", sep="")
  training <- data.frame(cbind(training, predvars))  
  
  for(j in baseJust) {
    for(i in baseJust) {
      if(!(i %in% swap)) {
        misclassrates[i,j] <- misclass.rate(bestNonBuggyTree(forms[[i]]))
      	if(criterion == "PRE") {
      	  goodPct <- 1 - misclassrates[i,j]
      	  margPct <- marginalPct[i]
      	  misclassrates[i,j] <- (goodPct - margPct) / (1 - margPct)
      	  misclassrates[i,j] <- misclassrates[i,j] * -1  # We want the *Largest* PRE
      	}
        cat("."); flush()
      }
    }
    z <- order(misclassrates[,j])[1]
    ## Note that in the case of ties, this gives the lower index number...
    
##    cat("\nMinimum in column ", j, " is row ", z, justnames[z], " with value ", misclassrates[z, j])
##    flush()

    swap[j] <- z
    if(method == "predicted") {
      predtree <- bestNonBuggyTree(forms[[z]])
      training[ ,z+offset] <- predtree$frame$yval[predtree$where]
      attach(training)
    }
    # Now update formulas!
    for(i in baseJust) {
      if(!(i %in% swap)) {
        newform <- add.term(forms[[i]], training, swap + offset)
        forms[[i]][3] <- newform[3]
      }
    }
  }
  
  misclassrates <- as.data.frame(round(misclassrates, 3))
  row.names(misclassrates) <- jDirVars
  formsColl <- forms
  
  for(i in 1:9) {
	assign(paste("tree.", i, sep=""), bestNonBuggyTree(forms[[i]]))
  }
  treesColl <- list(tree.1, tree.2, tree.3, tree.4, tree.5, tree.6, tree.7, tree.8, tree.9)
  names(treesColl) <- paste(justnames, "tree", sep="")	
  
  # Now-- How many justices, on average, are misclassified on each case?
  training$jmiscount <- justiceMisclassCount(treesColl)
 
  swap <- swap[!is.na(swap)]
  cat("\n\nJustice order: ", justnames[swap], "\n"); flush()

  stree <- predict.tree(treesColl[[swap[1]]], testing, type="tree")
  testing.trees <- list(stree, stree, stree, stree, stree, stree, stree, stree, stree)
  j <- 1
  for(i in swap) {
    attach(testing)
    thiscol <- paste(justnames[i], "dir", sep="")
##    checkpoint(paste("Predicting on test sample for ", justnames[i]))
    newvar <- paste(justnames[i],"predTree", sep ="")
    testing.trees[[j]] <- predict.tree(treesColl[[i]],testing,type="tree")
    testing.trees[[j]]$y <- testing[, which(names(testing) == as.character(formsColl[[i]][2]))]
    testing <- cbind(testing, as.data.frame(testing.trees[[j]]$frame$yval[testing.trees[[j]]$where]))
    names(testing)[ncol(testing)] <- paste(justnames[i], "pred", sep="")
    j <- j + 1
  }
  names(testing.trees) <- paste(justnames[swap], "tree", sep="")

  training$courtdir <- getCourtdir(training, method="dir")
  training$courtpred <- getCourtdir(training, method="pred")

##  cat("\nCreating testout object...", nrow(testing), "rows");flush()
  
  ## predicted unanimous is always 0 in the mixed function... guarantees we don't list a case
  ## as predicted to be both unanCon and unanLib
  testing$unanLibPred <- rep(0, nrow(testing))
  testing$unanConPred <- rep(0, nrow(testing))

  attach(testing)
  docket <- as.character(testing$docket)
  justdirs <- getJDirVarNames()
  justpreds <- getJPredVarNames()

  uLibPred <- testing$unanLibPred
  uConPred <- testing$unanConPred

  sVotes <- sumVotes(testing, justdirs)
  jLibAct <- sVotes$Zeros
  jConAct <- sVotes$Ones
  cDirAct <- getCourtdir(testing, method="dir")

  sVotes <- sumVotes(testing, justpreds)
  jLibPred <- sVotes$Zeros
  jConPred <- sVotes$Ones
  cDirPred <- getCourtdir(testing, method="pred")
  testing$unanLibPred <- (sVotes$Zeros == 9)
  testing$unanConPred <- (sVotes$Ones == 9)

  
  jMiscount <- justiceMisclassCountDF(testing)
  cMisclass <- as.integer(cDirPred != cDirAct)
   
  testout <- as.data.frame(cbind(docket, uConPred, uLibPred, jConAct, jConPred, jLibAct, jLibPred, cDirAct, cDirPred, jMiscount, cMisclass))
  
  ## cat("\nCreating object returned by predictedMixed()");flush()
  retObject <- list(justdirs, training, testing, misclassrates, formsColl, 
    treesColl, testing.trees, testout)
  ## checkpoint("\nretObject built...")
  names(retObject) <- c("justnames", "training", "testing", "misclassrates", 
    "formulae", "training.trees", "testing.trees", "testout")
  ## checkpoint("\nnames(retObject) assigned...")
  ## cat("\n")
  return(retObject)
}

bestNonBuggyTree <- function(formula, mdev=0.010, quiet=TRUE, ...) {
  ## Examines trees, beginning with default mindev value.
  ## Some trees can be built with deviance values of NaN in nodes;
  ## these trees cause R to crash when used to predict.
  ## This function returns the tree with the smallest mindev
  ## greater than or equal to the starting mindev value for which
  ## no deviances are NaN.
  require(tree)
  testTree <- tree(formula, na.action = na.pass, mindev=mdev, ...)
  while(any(is.na(testTree$frame$dev))) {
    mdev <- mdev + 0.001
    testTree <- tree(formula, na.action=na.pass, mindev = mdev, ...)
  }
  if(!quiet) { cat(as.character(formula[2]), " bestNonBuggyTree used with mindev =", mdev, "\n") }
  testTree
}

##
## End of MAJOR functions
##
################################################################################



################################################################################
##
## VARIABLE NAMES functions start here
##

getJusticeNames <- function(...) {
  return(c("rehn", "stev", "ocon", "scal", "ken", "sout", "thom", "gin", "bry"))
}

getIDVarNames <- function(...) {
  return(c("docket", "term"))
}

getJDirVarNames <- function(...) {
  jnames <- getJusticeNames()
  return(paste(jnames, "dir", sep=""))
}

getJPredVarNames <- function(...) {
  jnames <- getJusticeNames()
  return(paste(jnames, "pred", sep=""))
}

getPredictors <- function(combineEntities, ...) {
  retPreds <- c("lctdir", "value", "circnum", "petitnum", "respnum", "jur", "lcdisag", "statesc", "unconst", "otprec")
  if(combineEntities) {
    retPreds <- setdiff(retPreds, c("petitnum", "respnum"))
    retPreds <- c(retPreds, "prcode")
  }
  return(retPreds)
}

getFileVarNames <- function(...) {
  retVars <- getPredictors(...)
  notExpected <- c("value", "circnum", "petitnum", "respnum", "prcode")
  toDeriveFrom <- c("issue", "circuit", "petit", "respon")
  retVars <- setdiff(retVars, notExpected)
  retVars <- c(retVars, toDeriveFrom)
  return(retVars)
}

##
##  End of VARIABLE NAMES functions
##
################################################################################


################################################################################
##
##  PREP AUX functions start here
##

combinePetitResp <- function(pnums, rnums, ...) {
  pnums <- as.integer(as.character(pnums))
  rnums <- as.integer(as.character(rnums))
  pnums[is.na(pnums)] <- 0
  rnums[is.na(rnums)] <- 0
  A = pmax(pnums, rnums)
  B = pmin(pnums, rnums)
  N = A * 100 + B
  N[N==0] <- NA
  return(N)
}

convertCircuit <- function(vector, ...) {
  ## Takes a vector of string values with circuit labels and 
  ## returns same-length factor of numeric labels
  returnVec <- rep(NA, length(vector))
  returnVec[as.character(vector) == "1st"] <- 1
  returnVec[as.character(vector) == "2nd"] <- 2
  returnVec[as.character(vector) == "3rd"] <- 3
  returnVec[as.character(vector) == "4th"] <- 4
  returnVec[as.character(vector) == "5th"] <- 5
  returnVec[as.character(vector) == "6th"] <- 6
  returnVec[as.character(vector) == "7th"] <- 7
  returnVec[as.character(vector) == "8th"] <- 8
  returnVec[as.character(vector) == "9th"] <- 9
  returnVec[as.character(vector) == "10th"] <- 10
  returnVec[as.character(vector) == "11th"] <- 11
  returnVec[as.character(vector) == "DC"] <- 12
  returnVec[as.character(vector) == "FED"] <- 13
  returnVec <- factor(returnVec, levels=c(as.character(1:13)))
  return(returnVec)
}

convertEntity <- function(vector, condenseEntities=0, ...) {
  ## Takes a vector of string values with petitioner/respondent
  ## labels and returns a same-length factor of numeric labels
  returnVec <- rep(NA, length(vector))
  checkpoint(paste("convertEntity thinks condenseEntities = ", condenseEntities))
  if(condenseEntities == 0) {  ## original 11 factors
    returnVec[as.character(vector) == "BUSINESS"] <- 1
    returnVec[as.character(vector) == "CITY"]     <- 2
    returnVec[as.character(vector) == "DEF"]      <- 3
    returnVec[as.character(vector) == "EE"]       <- 4
    returnVec[as.character(vector) == "ER"]       <- 5
    returnVec[as.character(vector) == "INDIAN"]   <- 6
    returnVec[as.character(vector) == "IP"]       <- 7
    returnVec[as.character(vector) == "OF"]       <- 8
    returnVec[as.character(vector) == "POL"]      <- 9
    returnVec[as.character(vector) == "STATE"]    <- 10
    returnVec[as.character(vector) == "US"]       <- 11
    returnVec <- factor(returnVec, levels=c(as.character(1:11)))
  } else {
    returnVec[as.character(vector) == "BUSINESS"] <- 1
    returnVec[as.character(vector) == "ER"]       <- 1
    returnVec[as.character(vector) == "CITY"]     <- 2
    returnVec[as.character(vector) == "STATE"]    <- 2
    returnVec[as.character(vector) == "DEF"]      <- 3
    returnVec[as.character(vector) == "INDIAN"]   <- 3
    returnVec[as.character(vector) == "IP"]       <- 3
    returnVec[as.character(vector) == "EE"]       <- 4
    returnVec[as.character(vector) == "OF"]       <- 5
    returnVec[as.character(vector) == "POL"]      <- 5
    returnVec[as.character(vector) == "US"]       <- 6
    returnVec <- factor(returnVec, levels=c(as.character(1:6)))
  }
  return(returnVec)
}

spaethValue <- function(issue, ...) {
  ## Returns a list of same length as issue, containing
  ## Spaeth VALUE codes.  issue is assumed to contain
  ## Spaeth ISSUE codes.  0 is coded for unrecognized values.
  value <- rep(0, length(issue))
  value[issue >=  10 & issue <= 199] <- 1   # Criminal procedure
  value[issue >= 210 & issue <= 399] <- 2   # Civil rights
  value[issue >= 401 & issue <= 472] <- 3   # First amendment
  value[issue >= 501 & issue <= 507] <- 4   # Due process
  value[issue >= 531 & issue <= 537] <- 5   # Privacy
  value[issue >= 542 & issue <= 548] <- 6   # Attorneys
  value[issue >= 553 & issue <= 599] <- 7   # Unions
  value[issue >= 601 & issue <= 699] <- 8   # Economic activity
  value[issue >= 701 & issue <= 899] <- 9   # Judicial power
  value[issue >= 900 & issue <= 949] <- 10  # Federalism
  value[issue >= 950 & issue <= 959] <- 11  # Interstate relations
  value[issue >= 960 & issue <= 979] <- 12  # Federal taxation
  value[issue >= 980 & issue <= 989] <- 13  # Miscellaneous
  value <- factor(value, levels=c(as.character(1:13)))
  return(value)
}

##
##  End of PREP AUX functions
##
################################################################################


################################################################################
##
##  COUNTING functions start here
##

sumVotes <- function(dataframe, varnames, ...) {

  attach(dataframe)
  Ones <- rep(0, nrow(dataframe))
  Zeros <- rep(0, nrow(dataframe))
  NACount <- rep(0, nrow(dataframe))
  
  
  ## Now add 1 for each justice voting each way...
  for(var in varnames) {
    tvar <- as.integer(as.character(eval(as.name(var))))
    Ones <- Ones + ifelse(!is.na(tvar), as.integer(tvar == 1), 0)
    Zeros <- Zeros + ifelse(!is.na(tvar), as.integer(tvar == 0), 0)
    NACount <- NACount + as.integer(is.na(tvar))
  }
  
  ## BUT... if all the variables were NA, then both columns should be NA
  NACount <- Ones + Zeros
  Ones[NACount == 0] <- NA
  Zeros[NACount == 0] <- NA
  
  retdf <- data.frame(Ones, Zeros)
   
  return(retdf)
}

justiceMisclassCount <- function(tree.list) {
  checkpoint("In justiceMisclassCount()...")
  ## Function assumes tree.list has a list of trees all built on a common set of values.
  ## So we're assuming a common length of the $y component across the trees.
  ## Weird things will occur if that's not the case, so let's make sure to check it out...

  baseLen <- length(tree.list[[1]]$y)
  for(i in 2:length(tree.list)) {
    if(length(tree.list[[i]]$y) != baseLen) {
      stop("In justiceMisclassCount, at least one tree in the list is a different length than the others.")
    }
  }

  jMisclass <- rep(0, length(tree.list[[1]]$y))
  for(i in 1:length(tree.list)) {
    # A misclassification occurs when the justice's observed direction on a case ($y)
    # doesn't match the yvalue of the leaf where that case wound up
    jMisclass <- jMisclass + as.integer(tree.list[[i]]$y != tree.list[[i]]$frame$yval[tree.list[[i]]$where])
  }
  checkpoint("Going home from justiceMisclassCount()")
  return(jMisclass)
}

justiceMisclassCountDF <- function(dataframe) {
  ## Expects to find individual justice dir and pred variables.
  ## Returns list of NAs if the variables are not available.
  
  jDirVars <- getJDirVarNames()
  jPredVars <- getJPredVarNames()
  
  a <- setdiff(jDirVars, names(dataframe))
  b <- setdiff(jPredVars, names(dataframe))
  if(length(a) > 0 | length(b) > 0) {
    jMisclass <- rep(NA, nrow(dataframe))
    cat("\njusticeMisclassCountDF() is missing ", a, b)
    return(jMisclass)
  }
  
  cat("\njusticeMisclassCountDF(): All dir and pred variables present in dataframe")
  
  attach(dataframe)
  jMisclass <- rep(0, nrow(dataframe))
  
  for(i in 1:length(jDirVars)) {
    dir <- as.integer(as.character(eval(as.name(jDirVars[i]))))
    pred <- as.integer(as.character(eval(as.name(jPredVars[i]))))
    for(j in 1:nrow(dataframe)) {
      if(!is.na(dir[j]) & !is.na(pred[j]) & dir[j] != -1) {
        jMisclass[j] = jMisclass[j] + as.integer(dir[j] != pred[j])
      }
    }
  }
  
  cat("\nDone with justiceMisclassCountDF()...")
  cat("\n")
  return(jMisclass)
}

getCourtdir <- function(dataframe, method=c("dir", "pred")) {
  ## Assumes rehndir, stevdir, etc., exist in dataframe
  ## Or, for method="pred", uses rehnpred, stevpred, etc.
  attach(dataframe)

  method <- match.arg(method)
  varnames <- getJDirVarNames()
  if(method=="pred") varnames <- getJPredVarNames()
  
  missingvars <- setdiff(varnames, names(dataframe))
  if(length(missingvars) > 0) {
    cat(paste("\nIn getCourtdir, dataframe", deparse(substitute(dataframe)), 
      "\ndoes not contain variables", list(missingvars), "\nfor method", method))
    cat("\nReturning a vector of NAs\n")
    flush()
    return(rep(NA, nrow(dataframe)))
  }
  
  sVotes <- sumVotes(dataframe, varnames)
  attach(sVotes)
  courtdir <- as.integer(Ones > Zeros)
  courtdir[Ones == Zeros] <- NA

  return(courtdir)
}

misclass.rate <- function(tree) {
  ## Returns misclassification RATE for tree
  return(misclass.tree(tree)/tree$frame$n[1])
}

newjMat <- function(dataframe) {
 ############################################################
 ## Create a data frame "jMat" to include name, measure, error, and mean [valid] vote
 ## of each justice.
 ## These measures are based on BIGSTEPS run with 100% of cases.  Values may be
 ## slightly different if they were on 90% of cases... but according to IRT theory,
 ## they won't be different enough to matter.
 
 jName <- c("rehn", "stev",  "ocon", "scal", "ken",  "sout",  "thom", "gin",   "bry")
 jMeas <- c(842.07, 1240.30, 935.71, 822.35, 943.75, 1149.01, 795.56, 1147.92, 1123.35)
 jErr  <- c( 15.65,   16.88,  13.97,  16.27,  13.83,   14.70,  17.21,   14.64,   14.28)
 
 # Get mean justice vote among cases in which they voted either 0 or 1 [abstentions coded as 9]
 attach(dataframe)
 jMeanLib <- c((length(rehndir[rehndir==1])/length(rehndir[rehndir == 1 | rehndir == 0])),
               (length(stevdir[stevdir==1])/length(stevdir[stevdir == 1 | stevdir == 0])),
               (length(ocondir[ocondir==1])/length(ocondir[ocondir == 1 | ocondir == 0])),
               (length(scaldir[scaldir==1])/length(scaldir[scaldir == 1 | scaldir == 0])),
               (length( kendir[ kendir==1])/length( kendir[ kendir == 1 |  kendir == 0])),
               (length(soutdir[soutdir==1])/length(soutdir[soutdir == 1 | soutdir == 0])),
               (length(thomdir[thomdir==1])/length(thomdir[thomdir == 1 | thomdir == 0])),
               (length( gindir[ gindir==1])/length( gindir[ gindir == 1 |  gindir == 0])),
               (length( brydir[ brydir==1])/length( brydir[ brydir == 1 |  brydir == 0]))  )
 # For a benchmark, we need to do at least as well as the mean of the jutices' votes
 # or 1 minus the mean, if the mean is beow 0.5
 jBenchmark <- pmax(jMeanLib, 1-jMeanLib)
 jMat <- data.frame(jName, jMeas, jErr, jMeanLib, jBenchmark)
 rm(jName, jMeas, jErr, jMeanLib, jBenchmark)
 return(jMat)
 
 #   jName   jMeas  jErr  jMeanLib jBenchmark
 # 1  rehn  842.07 15.65 0.6700855  0.6700855
 # 2  stev 1240.30 16.88 0.3390706  0.6609294
 # 3  ocon  935.71 13.97 0.5958549  0.5958549
 # 4  scal  822.35 16.27 0.6850258  0.6850258
 # 5   ken  943.75 13.83 0.5890411  0.5890411
 # 6  sout 1149.01 14.70 0.4013722  0.5986278
 # 7  thom  795.56 17.21 0.7010309  0.7010309
 # 8   gin 1147.92 14.64 0.4023973  0.5976027
 # 9   bry 1123.35 14.28 0.4211438  0.5788562
                                                           ##
 ############################################################
}


##
##  End of COUNTING functions
##
################################################################################


################################################################################
##
##  FORUMLA SPECIFIC functions start here
##

baseform <- function(...) {
  # Provide a function as "dummydir ~ predictors" that relies on a function
  # that gives the names of the predictors
  predictors <- getPredictors(...)
  form <- dummydir ~ predictor1
  expr <- substitute(dummydir ~ predictor1, list(predictor1 = as.name(predictors[1])))
  form <- eval(expr)
  for(i in 2:length(predictors)) {
    expr <- substitute(update(form, . ~ . + newvar), list(newvar=as.name(predictors[i])))
    form <- eval(expr)
  }
  return(form)
}

justform <- function(formula, dirvarname) {
  expr <- substitute(update(formula, newjust ~ .), list(newjust =as.name(dirvarname)))
#  expr <- substitute(update(formula, as.factor(newjust) ~ .), list(newjust =as.name(dirvarname)))
  return(eval(expr))
}

add.term <- function(formula, matrix, swap) {
#  checkpoint("In add.term")
  for(i in swap) {
#    checkpoint(c("i = ", i))
    if(!is.na(i)) {
      at.local <- formula
      expr <- substitute(update(at.local, . ~ . + newvar), list(newvar= as.name(names(matrix)[i])))
      formula <- eval(expr)
    }
  }
#  checkpoint ("Leaving add.term")
  return(formula)
}

##
##  End of FORMULA SPECIFIC functions
##
################################################################################



################################################################################
##
##  AUXILIARY functions start here
##

VoteCode <- function() {
  retlist <- list(1,0)
  names(retlist) <- c("Conservative", "Liberal")
  return(retlist)
}

checkpoint <- function(string) {
  cat(string, "\n")
  ##if(is$windows) ## According to docs, flush.console() "does nothing under [non-Windows] front ends"
  ##  flush.console()  ## Commented out because, contrary to docs, this command just doesn't work under Unix
}

flush <- function() {
  ## if (is$windows) ## According to docs, flush.console() "does nothing under [non-Windows] front ends"
  ##  flush.console()  ## Commented out because, contrary to docs, this command just doesn't work under Unix
}


##
##  End of AUXILIARY functions
##
################################################################################
